/* $Id: submit.c 533 2007-12-22 15:25:42Z lukasz $ */
/*
 *  FedStage DRMAA for PBS Pro
 *  Copyright (C) 2006-2007  Fedstage Systems Inc.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file submit.c
 * Submitting job to PBS
 */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>

#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include <drmaa_utils/attrib.h>
#include <drmaa_utils/compat.h>
#include <drmaa_utils/conf.h>
#include <drmaa_utils/drmaa_impl.h>
#include <drmaa_utils/util.h>
#include <drmaa_utils/datetime.h>
#include <pbs_drmaa/pbs_attrib.h>
#include <pbs_drmaa/pbs_drmaa.h>

#ifndef lint
static char rcsid[]
#	ifdef __GNUC__
		__attribute__ ((unused))
#	endif
	= "$Id: submit.c 533 2007-12-22 15:25:42Z lukasz $";
#endif


void
drmaa_run_job_impl(
		char *job_id, size_t job_id_len,
		const drmaa_job_template_t *jt, int bulk_no,
		drmaa_err_ctx_t *err
		)
{
	drmaa_session_t *session      = NULL;
	drmaa_submit_ctx_t *ctx = NULL;
	char *destination = NULL;
	char *pbs_job_id;
	drmaa_job_t *job = NULL;

	DEBUG(( "-> drmaa_run_job_impl( bulk_no=%d )", bulk_no ));
	ctx = drmaa_create_submission_context( jt, bulk_no, err );

	drmaa_mutex_lock( (drmaa_mutex_t*)&jt->mutex, err );
	if( OK(err) )
		drmaa_apply_native_spec( ctx, err );
	if( OK(err) )
		drmaa_create_job_script( ctx, err );
	if( OK(err) )
		drmaa_set_job_submit_state( ctx, err );
	if( OK(err) )
		drmaa_set_job_files( ctx, err );
	if( OK(err) )
		drmaa_set_file_staging( ctx, err );
	if( OK(err) )
		drmaa_set_job_resources( ctx, err );
	if( OK(err) )
		drmaa_set_job_environment( ctx, err );
	if( OK(err) )
		drmaa_set_job_email_notif( ctx, err );
	if( OK(err) )
		drmaa_apply_job_category( ctx, err );
	if( OK(err) )
		drmaa_set_job_defaults( ctx, err );
	if( OK(err) )
	 {
		DRMAA_MALLOC( job, drmaa_job_t );
		drmaa_job_init( job, err );
	 }
	drmaa_mutex_unlock( (drmaa_mutex_t*)&jt->mutex, err );

	/* check for error */
	if( !OK(err) )
	 {
		drmaa_free_submission_context( ctx, err );
		return;
	 }

	if( OK(err) )
	 {
		struct attrl *i, **p;
		p = & ctx->impl->pbs_attribs;
		for( i = *p;  i != NULL;  i = i->next )
		 {
			if( !strcmp( i->name, "destination" ) )
			 {
				destination = i->value;
				*p = i->next;
				DRMAA_FREE( i );
				break;
			 }
			p = & i->next;
		 }

		if( destination == NULL )
			destination = drmaa_strdup( "",  err );
	 }

#if DRMAA_DEBUG
	if( OK(err) )
		drmaa_log_attrl( ctx->impl->pbs_attribs, "job attr:" );
#endif

	if( OK(err) )
	 {
		session	= jt->session;
		job->submit_time = time(NULL);
		drmaa_mutex_lock( &session->impl->conn_mutex, err );
		pbs_job_id = pbs_submit(
				session->impl->pbs_conn,
				(struct attropl*)ctx->impl->pbs_attribs,
				ctx->impl->script_filename,
				destination, NULL );
		DEBUG((
					"pbs_submit( fd=%d, attribs={...}, script=%s, destination=%s ) =%s",
					session->impl->pbs_conn, ctx->impl->script_filename, destination,
					pbs_job_id ));
		drmaa_mutex_unlock( &session->impl->conn_mutex, err );
		drmaa_free_submission_context( ctx, err );
	 }

	DRMAA_FREE( destination );

	if( OK(err) )
	 {
		if( pbs_job_id == NULL )
			RAISE_PBS();
		else
		 {
			const char *submit_state = jt->attrib[ ATTR_JOB_SUBMIT_STATE ];
			strlcpy( job_id, pbs_job_id, job_id_len );
			job->session = session;
			job->job_id = pbs_job_id;
			if( submit_state != NULL
					&&  !strcmp( submit_state, DRMAA_SUBMISSION_STATE_HOLD ) )
				job->flags |= DRMAA_JOB_HOLD;
		 }
	 }

	if( OK(err) )
		drmaa_job_set_add( session->jobs, job, err );
	else if( job != NULL )
	 {
		job->ref_cnt = 1;
		drmaa_job_release( &job, err );
	 }

	DEBUG(( "<- drmaa_run_job_impl; job_id=%s", OK(err) ? job_id : "<none>" ));
}


drmaa_submit_ctx_t *
drmaa_create_submission_context(
		const drmaa_job_template_t *jt, int bulk_no,
		drmaa_err_ctx_t *err
		)
{
	drmaa_submit_ctx_t *ctx;
	DRMAA_MALLOC( ctx, drmaa_submit_ctx_t );
	if( OK(err) )
	 {
		ctx->jt = jt;
		ctx->home_directory = NULL;
		ctx->working_directory = NULL;
		ctx->bulk_incr = NULL;
		DRMAA_MALLOC( ctx->impl, drmaa_submit_impl_t );
	 }
	if( OK(err) )
	 {
		drmaa_submit_impl_t *impl = ctx->impl;
		impl->pbs_attribs = NULL;
		memset( impl->pbs_attribs_bitset, 0, sizeof(impl->pbs_attribs_bitset) );
		impl->script_filename = NULL;
		ctx->home_directory = drmaa_strdup( getenv("HOME"), err );
		if( ctx->home_directory == NULL )
			ctx->home_directory = drmaa_strdup( "/", err );
		if( jt->attrib[ ATTR_JOB_WORKING_DIR ] != NULL )
			ctx->working_directory = drmaa_strdup(
					jt->attrib[ ATTR_JOB_WORKING_DIR ], err );
		else
			ctx->working_directory = drmaa_strdup( ctx->home_directory, err );
		ctx->bulk_incr = drmaa_asprintf( err, "%d", bulk_no );
	 }

	if( OK(err) )
		return ctx;
	else
	 {
		drmaa_free_submission_context( ctx, err );
		return NULL;
	 }
}


void
drmaa_free_submission_context( drmaa_submit_ctx_t *ctx,
	 drmaa_err_ctx_t *err )
{
	if( ctx == NULL )
		return;
	if( ctx->impl != NULL )
	 {
		if( ctx->impl->script_filename != NULL )
		 {
			unlink( ctx->impl->script_filename );
			DRMAA_FREE( ctx->impl->script_filename );
		 }
		drmaa_free_attrl( ctx->impl->pbs_attribs );
		DRMAA_FREE( ctx->impl );
	 }
	DRMAA_FREE( ctx->working_directory );
	DRMAA_FREE( ctx->home_directory );
	DRMAA_FREE( ctx->bulk_incr );
	DRMAA_FREE( ctx );
}



struct drmaa_def_attr_s {
	int         attr;
	const char *value;
};

void
drmaa_set_job_defaults(
		drmaa_submit_ctx_t *ctx,
		drmaa_err_ctx_t *err
		)
{
	static const struct drmaa_def_attr_s tab[] = {
	 { ATTR_CHECKPOINT, "u" },
	 { ATTR_KEEP_FILES, "n" },
	 { ATTR_PRIORITY,   "0" },
	 { ATTR_RERUNABLE,  "y" },
	 { ATTR_SHELL,      "/bin/sh" },
	 { -1, NULL } };
	const struct drmaa_def_attr_s *i;

	for( i = tab;  i->attr != -1 && OK(err);  i++ )
		drmaa_add_pbs_attr( ctx, i->attr, NULL, strdup(i->value),
				0, err );
}


void
drmaa_set_job_submit_state(
		drmaa_submit_ctx_t *ctx,
		drmaa_err_ctx_t *err
		)
{
	void **attrib = ctx->jt->attrib;
	const char *submit_state = NULL;
	const char *drmaa_start_time = NULL;
	const char *hold_types = NULL;
	const char *job_name = NULL;

	if( !OK(err) )
		return;

	submit_state = attrib[ ATTR_JOB_SUBMIT_STATE ];
	drmaa_start_time = attrib[ ATTR_START_TIME ];
	job_name = attrib[ ATTR_JOB_NAME ];

	if( job_name != NULL )
	 {
		drmaa_add_pbs_attr( ctx,
				ATTR_JOB_NAME, NULL, drmaa_strdup( job_name, err ),
				DRMAA_PLACEHOLDER_MASK_INCR, err );
	 }

	if( submit_state == NULL || !strcmp(submit_state,DRMAA_SUBMISSION_STATE_ACTIVE) )
		hold_types = "n";
	else if( !strcmp(submit_state,DRMAA_SUBMISSION_STATE_HOLD) )
		hold_types = "u";
	else
		RAISE_DRMAA_1( DRMAA_ERRNO_INVALID_ATTRIBUTE_VALUE );

	if( OK(err) )
		drmaa_add_pbs_attr( ctx, ATTR_HOLD_TYPES, NULL, strdup(hold_types),
			0, err );
	if( OK(err)  &&  drmaa_start_time != NULL )
	 {
		char pbs_start_time[20];
		time_t start_time;
		struct tm start_time_tm;
		start_time = drmaa_parse_datetime( drmaa_start_time, err );
		if( OK(err) )
		 {
			localtime_r( &start_time, &start_time_tm );
			sprintf( pbs_start_time, "%04d%02d%02d%02d%02d.%02d",
					start_time_tm.tm_year + 1900,
					start_time_tm.tm_mon + 1,
					start_time_tm.tm_mday,
					start_time_tm.tm_hour,
					start_time_tm.tm_min,
					start_time_tm.tm_sec
					);
			drmaa_add_pbs_attr( ctx, ATTR_START_TIME, NULL,
					drmaa_strdup( pbs_start_time, err ), 0, err );
		 }
	 }
}


/** Writes job's PBS script to temporary file. */
void
drmaa_create_job_script(
		drmaa_submit_ctx_t *ctx,
		drmaa_err_ctx_t *err
		)
{
	size_t script_size = 0;
	const char *executable = NULL;
	const char *wd = NULL;
	const char **argv = NULL;
	char *input_path = 0;
	char *script = NULL;
	char *s;
	const char **i;
	void **attrib = ctx->jt->attrib;

	executable   = (const char*) attrib[ ATTR_JOB_PATH        ];
	wd           = (const char*) attrib[ ATTR_JOB_WORKING_DIR ];
	argv         = (const char**)attrib[ ATTR_ARGV            ];
	input_path   =               attrib[ ATTR_INPUT_PATH      ];

	if( executable == NULL )
		RAISE_DRMAA_1( DRMAA_ERRNO_INVALID_ATTRIBUTE_VALUE );

	if( OK(err)  &&  input_path != NULL )
	 {
		char *colon = strchr( input_path, ':' );
		if( colon != NULL )
			input_path = colon+1;
	 }

	if( OK(err) )
	 { /* compute script length */
		script_size = 0;
		if( wd != NULL )
			script_size += strlen("cd ") + strlen(wd) + strlen("; ");
		script_size += strlen("exec ") + strlen(executable);
		if( argv != NULL )
			for( i = argv;  *i != NULL;  i++ )
				script_size += 1+strlen(*i);
		if( input_path != NULL )
			script_size += strlen(" <") + strlen(input_path);
		/* script_size += 1; */
	 }

	if( OK(err) )
		DRMAA_CALLOC( script, script_size+1, char );
	if( OK(err) )
	 {
		s = script;
		if( wd != NULL )
			s += sprintf( s, "cd %s; ", wd );
		s += sprintf( s, "exec %s", executable );
		if( argv != NULL )
			for( i = argv;  *i != NULL;  i++ )
				s += sprintf( s, " %s", *i );
		if( input_path != NULL )
			s += sprintf( s, " <%s", input_path );
		/* *s++ = '\n'; */
		assert( s == script+script_size );
	 }

	if( OK(err) )
	 {
		DEBUG(( "script before substitution = {%s}", script ));
		script = drmaa_expand_placeholders( ctx, script,
				DRMAA_PLACEHOLDER_MASK_HD |
				DRMAA_PLACEHOLDER_MASK_WD |
				DRMAA_PLACEHOLDER_MASK_INCR, err );
		DEBUG(( "script after substitution = {%s}", script ));
	 }

	if( OK(err) )
		ctx->impl->script_filename
			= drmaa_write_tmpfile( script, strlen(script), err );

	DRMAA_FREE( script );
}


/**
 * Set files where job's standard input/output/error
 * streams will be redirected.
 */
void
drmaa_set_job_files(
		drmaa_submit_ctx_t *ctx,
		drmaa_err_ctx_t *err
		)
{
	void **attrib = ctx->jt->attrib;
	const char *join_files = attrib[ATTR_JOIN_FILES];
	bool b_join_files;
	int i;

	for( i = 0;  i < 2 && OK(err);  i++ )
	 {
		int attr = i ? ATTR_OUTPUT_PATH : ATTR_ERROR_PATH;
		char *path = (char*)attrib[attr];
		if( path != NULL )
		 {
			char *colon;
			colon = strchr( path, ':' );
			if( colon != NULL )
				path = colon+1;
			drmaa_add_pbs_attr( ctx, attr, NULL, drmaa_strdup(path, err),
				 DRMAA_PLACEHOLDER_MASK_HD
				 | DRMAA_PLACEHOLDER_MASK_WD
				 | DRMAA_PLACEHOLDER_MASK_INCR, err
				 );
		 }
	 }

	if( OK(err) )
	 {
		b_join_files = join_files != NULL  &&  !strcmp(join_files,"1");
		drmaa_add_pbs_attr( ctx,
				ATTR_JOIN_FILES, NULL, drmaa_strdup(b_join_files ? "y" : "n", err),
				0, err );
	 }
}


void
drmaa_set_file_staging(
		drmaa_submit_ctx_t *ctx,
		drmaa_err_ctx_t *err
		)
{
	void **attrib = NULL;
	char *transfer_files = NULL;
	char *input_file  = NULL;
	char *output_file = NULL;
	char *error_file  = NULL;
	char *output_staging = NULL;
	char *error_staging  = NULL;
	char *stageout =NULL;

	if( !OK(err) )
		return;

	attrib = ctx->jt->attrib;
	transfer_files = (char*)attrib[ ATTR_TRANSFER_FILES ];

	if( transfer_files )
	 {
		char *i;
		for( i = transfer_files;  *i;  i++ )
			switch( *i )
			 {
				case 'i':  input_file  = (char*)attrib[ ATTR_INPUT_PATH  ];  break;
				case 'o':  output_file = (char*)attrib[ ATTR_OUTPUT_PATH ];  break;
				case 'e':  error_file  = (char*)attrib[ ATTR_ERROR_PATH  ];  break;
				default:
					RAISE_DRMAA_1( DRMAA_ERRNO_INVALID_ATTRIBUTE_VALUE );
			 }
	 }
	else
		return;

	if( OK(err)  &&  input_file )
		drmaa_add_pbs_attr( ctx, ATTR_STAGEIN, NULL,
					drmaa_translate_staging( input_file, err ),
					DRMAA_PLACEHOLDER_MASK_HD |
					DRMAA_PLACEHOLDER_MASK_WD |
					DRMAA_PLACEHOLDER_MASK_INCR,
					err );

	if( OK(err)  &&  output_file )
		output_staging = drmaa_translate_staging( output_file, err );
	if( OK(err)  &&  error_file )
		error_staging = drmaa_translate_staging( error_file, err );

	if( OK(err) )
	 {
		if( output_staging && error_staging )
			stageout = drmaa_asprintf( err, "%s,%s", output_staging, error_staging );
		else if( output_staging )
			stageout = output_staging;
		else if( error_staging )
			stageout = error_staging;
	 }

	if( OK(err)  &&  stageout )
		drmaa_add_pbs_attr( ctx, ATTR_STAGEOUT, NULL, stageout,
					DRMAA_PLACEHOLDER_MASK_HD |
					DRMAA_PLACEHOLDER_MASK_WD |
					DRMAA_PLACEHOLDER_MASK_INCR,
					err );
	else if( stageout != NULL )
		DRMAA_FREE( stageout );

	if( output_staging  &&  output_staging != stageout )
		DRMAA_FREE( output_staging );
	if( error_staging  &&  error_staging != stageout )
		DRMAA_FREE( error_staging );
}


void
drmaa_set_job_resources(
		drmaa_submit_ctx_t *ctx,
		drmaa_err_ctx_t *err
		)
{
	const char *cpu_time_limit = NULL;
	const char *walltime_limit = NULL;

	if( !OK(err) )
		return;
	cpu_time_limit = (char*)ctx->jt->attrib[ ATTR_HARD_CPU_TIME_LIMIT ];
	walltime_limit = (char*)ctx->jt->attrib[ ATTR_HARD_WCT_LIMIT ];
	if( OK(err)  &&  cpu_time_limit )
		drmaa_add_pbs_attr(
				ctx, ATTR_RESOURCES,
				drmaa_strdup("pcput", err), drmaa_strdup(cpu_time_limit, err),
				0, err );
	if( OK(err)  &&  cpu_time_limit )
		drmaa_add_pbs_attr(
				ctx, ATTR_RESOURCES,
				drmaa_strdup("cput", err), drmaa_strdup(cpu_time_limit, err),
				0, err );
	if( OK(err)  &&  walltime_limit )
		drmaa_add_pbs_attr(
				ctx, ATTR_RESOURCES,
				drmaa_strdup("walltime", err),
				drmaa_strdup(walltime_limit, err),
				0, err );
}


/** Sets job environment variables. */
void
drmaa_set_job_environment(
		drmaa_submit_ctx_t *ctx,
		drmaa_err_ctx_t *err
		)
{
	void **attrib = NULL;
	size_t s;
	char *env;
	char *value = NULL;

	if( !OK(err) )
		return;

	attrib = ctx->jt->attrib;
	if( attrib[ATTR_ENV] == NULL )
		return;

	env = drmaa_strdup( "", err );
	if( !OK(err) )
		return;
	s = strlen( env );

	value = drmaa_explode( attrib[ATTR_ENV], ',', err );
	if( OK(err) )
		DRMAA_REALLOC( env, s + strlen(value) + 1, char );
	if( OK(err) )
	 {
		strcpy( env+s, value );
		DRMAA_FREE( value );
	 }
	drmaa_add_pbs_attr( ctx, ATTR_ENV, NULL, env,
			DRMAA_PLACEHOLDER_MASK_INCR, err );
}


/**
 * Set e-mail notification about job state.
 */
void
drmaa_set_job_email_notif(
		drmaa_submit_ctx_t *ctx,
		drmaa_err_ctx_t *err
		)
{
	void **attrib = NULL;
	const char *block_email = NULL;

	if( !OK(err) )
		return;

	attrib = ctx->jt->attrib;
	block_email = (const char*)attrib[ ATTR_BLOCK_EMAIL ];

	if( block_email != NULL  &&  !strcmp(block_email,"1") )
	 {
		drmaa_add_pbs_attr( ctx, ATTR_MAIL_POINTS, NULL, drmaa_strdup("n", err),
				0, err );
	 }
	else
	 {
		const char **mail_list = (const char**)attrib[ ATTR_EMAIL ];
		char *pbs_mail_list = NULL;
		if( mail_list != NULL )
		 {
			pbs_mail_list = drmaa_explode( mail_list, ',', err );
			if( OK(err) )
				drmaa_add_pbs_attr( ctx, ATTR_EMAIL, NULL, pbs_mail_list,
					0, err );
			if( OK(err) )
			 {
				/* a  mail is sent when the job is aborted by the batch system.
				   b  mail is sent when the job begins execution.
				   e  mail is sent when the job terminates. */
				drmaa_add_pbs_attr( ctx, ATTR_MAIL_POINTS, NULL, drmaa_strdup("ae", err),
						0, err );
			 }
			else
				DRMAA_FREE( pbs_mail_list );
		 }
	 }
}


void
drmaa_apply_native_spec(
		drmaa_submit_ctx_t *ctx, drmaa_err_ctx_t *err )
{
	const char *native = NULL;
	native = (const char*) ctx->jt->attrib[ ATTR_NATIVE ];
	if( OK(err)  &&  native != NULL )
		drmaa_parse_qsub_args( ctx, native, err );
}


void
drmaa_apply_job_category(
		drmaa_submit_ctx_t *ctx,
		drmaa_err_ctx_t *err
		)
{
	const char *job_category = NULL;
	const char *category_spec = NULL;
	drmaa_conf_option_t *value = NULL;

	if( OK(err) )
	 {
		job_category = (const char*) ctx->jt->attrib[ ATTR_JOB_CATEGORY ];
		if( job_category == NULL  ||  job_category[0]==0 )
			job_category = "default";
		value = drmaa_conf_dict_get( ctx->jt->session->job_categories,
				job_category, err );
		if( OK(err) &&  value  &&  value->type == DRMAA_CONF_STRING )
			category_spec = value->val.string;
	 }

	if( OK(err)  &&  category_spec != NULL )
		drmaa_parse_qsub_args( ctx, category_spec, err );
}



char *
drmaa_translate_staging( const char *stage, drmaa_err_ctx_t *err )
{
	char hostname[ HOST_NAME_MAX+1 ];
	const char *host = NULL, *filename = NULL;
	size_t hostlen = 0;
	char *result = NULL;
	char *p;

	if( !OK(err) )
		return NULL;

	p = strchr( (char*)stage, ':' );
	if( p == NULL )
	 {
		host = NULL;
		filename = stage;
	 }
	else
	 {
		if( p == stage )
			host = NULL;
		else
		 {
			host    = stage;
			hostlen = p - stage;
		 }
		filename = p + 1;
	 }
	if( host == NULL )
	 {
		gethostname( hostname, sizeof(hostname) );
		host = hostname;
		hostlen = strlen(hostname);
	 }
	result = drmaa_asprintf( err, "%s@%.*s:%s", err,
			filename, hostlen, host, filename );
	return result;
}


typedef enum{ NOP_, SET_, APPEND_ } attr_op_t;

/**
 * Adds jobs PBS attributes list.
 * @param attr_code  Attribute code (@ref drmaa_attribute_t).
 * @param value  Attribute value -- null terminated string allocated with
 *  malloc();  <b>callee</b> is responsible for freeing it.
 * @param placeholders  Placeholders to substitute inside value.
 *  It have following bits:
 *  - @ref DRMAA_PLACEHOLDER_MASK_HD -- replaces @ref DRMAA_PLACEHOLDER_HD with
 *    user's home direcotry,
 *  - @ref DRMAA_PLACEHOLDER_MASK_WD -- replaces @ref DRMAA_PLACEHOLDER_WD with
 *    job's working direcotry,
 *  - @ref DRMAA_PLACEHOLDER_MASK_INCR -- replaces @ref DRMAA_PLACEHOLDER_INCR
 *    with bulk job index.
 */
void
drmaa_add_pbs_attr(
		drmaa_submit_ctx_t *ctx,
		int attr_code, char *resource, char *value,
		unsigned placeholders,
		drmaa_err_ctx_t *err
		)
{
	struct attrl *attr = NULL;
	drmaa_session_t *session = ctx->jt->session;
	drmaa_attrib_info_t *attr_info = & session->attributes[ attr_code ];
	attr_op_t op = NOP_;

	if( !OK(err) )
		return;

	/*
	 * Take care of NULL values passed directly
	 * from failed call to strdup, asprintf, etc.
	 */
	if( value == NULL )
		RAISE_DRMAA_1( DRMAA_ERRNO_NO_MEMORY );

	if( OK(err) )
		value = drmaa_expand_placeholders( ctx, value, placeholders, err );

	if( OK(err) )
	 {
		uint32_t *p = ctx->impl->pbs_attribs_bitset + attr_code/32;
		uint32_t mask = 1 << (attr_code & 0x1f);
		if( *p & mask )
		 {
			switch( attr_code )
			 {
				case ATTR_ENV:
				case ATTR_RESOURCES: /* specialy treated */
				case ATTR_EMAIL:
				case ATTR_STAGEIN:
				case ATTR_STAGEOUT:
				case ATTR_DEPENDENCIES:
#if 0
				case ATTR_SHELL:
				case ATTR_USER_LIST:
				case ATTR_GROUP_LIST:
#endif
					op = APPEND_;
					break;
				default:
					op = NOP_;
					break;
			 }
		 }
		else
		 {
			*p |= mask;
			op = SET_;
		 }
	 }

	if( OK(err)  &&  op == APPEND_ )
	 {
		if( attr_code == ATTR_RESOURCES )
		 {
			struct attrl *i;
			op = SET_;
			for( i = ctx->impl->pbs_attribs;  i != NULL;  i = i->next )
				if( !strcmp( i->name, attr_info->drm_name )
						&& !strcmp( i->resource, resource ) )
					{
					 op = NOP_;
					 break;
					}
		 }
		else
		 {
			struct attrl *i;
			for( i = ctx->impl->pbs_attribs;  i != NULL;  i = i->next )
				if( !strcmp( i->name, attr_info->drm_name ) )
					break;
			assert( i != NULL );
			DEBUG(( "merging %s attribute (prev: '%s', new: '%s')",
						attr_info->drm_name, i->value, value ));
			DRMAA_REALLOC( i->value, strlen(i->value)+1+strlen(value)+1, char );
			if( OK(err) )
			 {
				size_t p = strlen( i->value );
				if( p > 0  &&  strlen(value) > 0 )
					i->value[ p++ ] = ',';
				strcpy( i->value + p, value );
			 }
		 }
	 }

	if( OK(err)  &&  op == SET_ )
	 {
		DRMAA_MALLOC( attr, struct attrl );

		if( OK(err) )
		 {
			attr->next     = ctx->impl->pbs_attribs;
			attr->name     = drmaa_strdup( attr_info->drm_name, err );
			attr->value    = value;
			attr->resource = resource;
			attr->op       = SET;
		 }

		if( OK(err) )
		 {
			ctx->impl->pbs_attribs = attr;
			attr = NULL;
			value = NULL;
			resource = NULL;
		 }

		DRMAA_FREE( attr );
	 }

	if( OK(err)  &&  op == NOP_ )
	 {
		DEBUG(( "ignoring attribute: %s=%s",
					attr_info->drm_name, value ));
	 }

	DRMAA_FREE( value );
	DRMAA_FREE( resource );
}


/**
 * Writes temporary file.
 * @param content   Buffer with content to write.
 * @param len       Buffer's length.
 * @return Path to temporary file upon successful return
 *  or NULL.
 */
char *
drmaa_write_tmpfile(
	 const char *content, size_t len,
	 drmaa_err_ctx_t *err
	 )
{
	char *name = NULL;
	int fd = -1;
#define TMPFILE_TEMPLATE "/tmp/pbs_drmaa.XXXXXX"

	DEBUG(( "-> drmaa_write_tmpfile" ));

	if( OK(err) )
	 {
		name = strdup( TMPFILE_TEMPLATE );
		if( name == NULL )
			RAISE_DRMAA_1( DRMAA_ERRNO_NO_MEMORY );
	 }

	if( OK(err) )
	 {
		fd = mkstemp( name );
		if( fd < 0 )
			RAISE_ERRNO();
	 }

	while( OK(err) && len > 0 )
	 {
		size_t written = write( fd, content, len );
		if( written != (size_t)-1 )
		 {
			content += written;
			len     -= written;
		 }
		else
		 {
			DRMAA_FREE( name );
			close( fd );
			RAISE_ERRNO();
		 }
	 }

	if( fd>=0 && close( fd ) )
		RAISE_ERRNO();

	if( !OK(err) )
	 {
		DRMAA_FREE( name );
		name = NULL;
	 }

	DEBUG(( "<- drmaa_write_tmpfile =%s", name ));
	return name;
}


void
drmaa_raise_qsub_parse_error(
		const char *message,
		const char *string, const char *pos,
		drmaa_err_ctx_t *err
		)
{
	char *errmsg = NULL;
	errmsg = drmaa_asprintf( err,
			"in native specification: %s:\n%s\n%*s^",
			message, string, pos-string, "" );
	if( errmsg )
		RAISE_DRMAA_2( DRMAA_ERRNO_INVALID_ATTRIBUTE_VALUE, errmsg );
	DRMAA_FREE( errmsg );
}


void
drmaa_parse_qsub_resources(
		drmaa_submit_ctx_t *ctx,
		char *resources_list,
		drmaa_err_ctx_t *err
		)
{
	const char *i = resources_list;

	while( OK(err)  &&  *i != '\0' )
	 {
		const char *name = i, *value = NULL;
		size_t name_len = 0, value_len = 0;

		while( *i != '='  &&  *i != ','  &&  *i != '\0' )
			i++;
		name_len = i - name;

		if( name_len == 0 )
		 {
			drmaa_raise_qsub_parse_error(
					"no resource name", resources_list, i, err );
			break;
		 }

		if( *i == '=' )
			i++;
		else
		 {
			drmaa_raise_qsub_parse_error(
				"no resource value", resources_list, i, err );
			break;
		 }

		value = i;
		while( *i != ','  &&  *i != '\0' )
			i++;
		value_len = i - value;

		drmaa_add_pbs_attr(
				ctx, ATTR_RESOURCES,
				drmaa_strndup( name, name_len, err ),
				drmaa_strndup( value, value_len, err ),
				0, err
				);

		if( *i == ',' )
			i++;
	 }

	DRMAA_FREE( resources_list );
}


typedef struct long_option_map_s {
	const char *optname;
	int attr_code;
} long_option_map_t;

void
drmaa_parse_qsub_args(
		drmaa_submit_ctx_t *ctx, const char *string,
		drmaa_err_ctx_t *err
		)
{
	enum {
		LONG_OPTION = -2,
		V_OPTION    = -3
	};
	static const int qsub_options_table[128] = {
		/* 0 - 31 */
			-1, -1, -1, -1, -1, -1, -1, -1,  -1, -1, -1, -1, -1, -1, -1, -1,
			-1, -1, -1, -1, -1, -1, -1, -1,  -1, -1, -1, -1, -1, -1, -1, -1,
		/* 32 - 39 */
			-1, -1, -1, -1, -1, -1, -1, -1,
		/* 40 - 47 */
			-1, -1, -1, -1, -1, /* `-' */ LONG_OPTION, -1, -1,
		/* 48 - 55 */
			-1, -1, -1, -1, -1, -1, -1, -1,
		/* 56 - 63 */
			-1, -1, -1, -1, -1, -1, -1, -1,
		/* 64 - 71 */ -1,
			/* A: Account_Name */ ATTR_ACCOUNT_NAME, -1, -1, -1, -1, -1, -1,
		/* 72 - 79 */
			-1, -1, -1, -1, -1,
			/* M: Mail_Users */ ATTR_EMAIL,
			/* N: Job_Name */ ATTR_JOB_NAME, -1,
		/* 80 - 87 */
			-1, -1, -1, /* S: Shell_Path_List */ ATTR_SHELL, -1, -1,
			/* -V */ V_OPTION, -1,
		/* 88 - 95 */
			-1, -1, -1, -1, -1, -1, -1, -1,
		/* 96 - 103 */ -1,
			/* a: Execution_Time */ ATTR_START_TIME, -1,
			/* c: Checkpoint */ ATTR_CHECKPOINT, -1,
			/* e: Error_Path */ ATTR_ERROR_PATH, -1,
			/* g: group_list */ ATTR_GROUP_LIST,
		/* 104 - 111 */
			/* h: Hold_Types */ ATTR_HOLD_TYPES, -1,
			/* j: Join_Path */ ATTR_JOIN_FILES,
			/* k: Keep_Files */ ATTR_KEEP_FILES,
			/* l: Resource_List */ ATTR_RESOURCES,
			/* m: Mail_Points */ ATTR_MAIL_POINTS, -1,
			/* o: Output_Path */ ATTR_OUTPUT_PATH,
		/* 112 - 119 */
			/* p: Priority */ ATTR_PRIORITY,
			/* q: destination */ ATTR_DESTINATION_QUEUE,
			/* r: Rerunable */ ATTR_RERUNABLE, -1, -1,
			/* u: User_List */ ATTR_USER_LIST,
			/* v: Variable_List */ ATTR_ENV, -1,
		/* 120 - 127 */
			-1, -1, -1, -1, -1, -1, -1, -1
	};
	static const long_option_map_t long_options[] = {
		 { "arid", ATTR_DESTINATION_QUEUE },
		 { "depend", ATTR_DEPENDENCIES },
		 { "group-list", ATTR_GROUP_LIST },
		 { "stagein", ATTR_STAGEIN },
		 { "stageout", ATTR_STAGEOUT },
		 { NULL, -1 },
	};
	const char *i = NULL;
	int attr_code;

	i = string;
	while( OK(err)  &&  *i != '\0' )
	 {
		while( *i == ' ' )
			i++;
		if( *i++ != '-' )
			drmaa_raise_qsub_parse_error( "`-' expected", string, i, err );
		if( !OK(err) )
			break;

		attr_code = qsub_options_table[ *(unsigned char*)i++ ];

		switch( attr_code )
		 {
			default:
			 {
				const char *value_begin = NULL;
				char *value = NULL;
				while( *i == ' ' )
					i++;
				/* Note: option values may start with `-' */
				if( *i == '\0' )
					drmaa_raise_qsub_parse_error(
						"no value for option", string, i, err );
				if( OK(err) )
				 {
					value_begin = i;
					while( *i != '\0'  &&  *i != ' ' )
						i++;
					value = drmaa_strndup( value_begin, i-value_begin, err );
				 }
				if( OK(err) )
				 {
					if( attr_code == ATTR_RESOURCES )
						drmaa_parse_qsub_resources( ctx, value, err );
					else
						drmaa_add_pbs_attr(
								ctx, attr_code, NULL, value,
								0, err
								);
				 }
				break;
			 }

			case ATTR_HOLD_TYPES:
				drmaa_add_pbs_attr(
						ctx, attr_code, NULL, drmaa_strdup( USER_HOLD, err ),
						0, err );
				break;

			case V_OPTION:
				drmaa_add_pbs_attr(
						ctx, ATTR_ENV, NULL,
							drmaa_explode( (const char*const*)environ, ',', err ),
						0, err );
				break;

			case LONG_OPTION:
			 {
				const char *optname;
				size_t optlen;
				const char *value_begin;
				char *value = NULL;
				const long_option_map_t *li;
				bool found = false;

				optname = i;
				while( isalnum( *(unsigned char*)i ) || *i == '-' )
					i++;
				optlen = i - optname;

				while( *i == ' ' )
					i++;
				if( *i == '=' )
					i++;
				value_begin = i;
				while( *i != '\0'  &&  *i != ' ' )
					i++;
				value = drmaa_strndup( value_begin, i-value_begin, err );

				for( li = long_options;  li->optname;  li++ )
					if( !strncmp( optname, li->optname, optlen )
							&&  optlen == strlen(li->optname) )
					 {
						drmaa_add_pbs_attr(
								ctx, li->attr_code, NULL, value, 0, err );
						found = true;
						value = NULL;
						break;
					 }

				if( OK(err)  &&  !found )
					drmaa_raise_qsub_parse_error(
						"invalid long option", string, i, err );
				DRMAA_FREE( value );
				break;
			 }

			case -1:
				drmaa_raise_qsub_parse_error( "invalid option", string, i, err );
				break;
		 }
	 }
}


