#
# $Id: drmaa.pyx,v 1.8 2007/03/12 18:11:13 ciesnik Exp $
#
# Python bindings to DRMAA
# Copyright (C) 2006-2007  Fedstage Systems Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import weakref


cdef enum:
	ERRLEN              = DRMAA_ERROR_STRING_BUFFER


# module constants

JOB_IDS_SESSION_ALL      = "DRMAA_JOB_IDS_SESSION_ALL"
JOB_IDS_SESSION_ANY      = "DRMAA_JOB_IDS_SESSION_ANY"
SUBMISSION_STATE_ACTIVE  = "drmaa_active"
SUBMISSION_STATE_HOLD    = "drmaa_hold"
PLACEHOLDER_HD           = "$drmaa_hd_ph$"
PLACEHOLDER_WD           = "$drmaa_wd_ph$"
PLACEHOLDER_INCR         = "$drmaa_incr_ph$"

SUSPEND    = DRMAA_CONTROL_SUSPEND
RESUME     = DRMAA_CONTROL_RESUME
HOLD       = DRMAA_CONTROL_HOLD
RELEASE    = DRMAA_CONTROL_RELEASE
TERMINATE  = DRMAA_CONTROL_TERMINATE
ERRNO_SUCCESS                            = DRMAA_ERRNO_SUCCESS
ERRNO_INTERNAL_ERROR                     = DRMAA_ERRNO_INTERNAL_ERROR
ERRNO_DRM_COMMUNICATION_FAILURE          = DRMAA_ERRNO_DRM_COMMUNICATION_FAILURE
ERRNO_AUTH_FAILURE                       = DRMAA_ERRNO_AUTH_FAILURE
ERRNO_INVALID_ARGUMENT                   = DRMAA_ERRNO_INVALID_ARGUMENT
ERRNO_NO_ACTIVE_SESSION                  = DRMAA_ERRNO_NO_ACTIVE_SESSION
ERRNO_NO_MEMORY                          = DRMAA_ERRNO_NO_MEMORY
ERRNO_INVALID_CONTACT_STRING             = DRMAA_ERRNO_INVALID_CONTACT_STRING
ERRNO_DEFAULT_CONTACT_STRING_ERROR       = DRMAA_ERRNO_DEFAULT_CONTACT_STRING_ERROR
ERRNO_NO_DEFAULT_CONTACT_STRING_SELECTED = DRMAA_ERRNO_NO_DEFAULT_CONTACT_STRING_SELECTED
ERRNO_DRMS_INIT_FAILED                   = DRMAA_ERRNO_DRMS_INIT_FAILED
ERRNO_ALREADY_ACTIVE_SESSION             = DRMAA_ERRNO_ALREADY_ACTIVE_SESSION
ERRNO_DRMS_EXIT_ERROR                    = DRMAA_ERRNO_DRMS_EXIT_ERROR
ERRNO_INVALID_ATTRIBUTE_FORMAT           = DRMAA_ERRNO_INVALID_ATTRIBUTE_FORMAT
ERRNO_INVALID_ATTRIBUTE_VALUE            = DRMAA_ERRNO_INVALID_ATTRIBUTE_VALUE
ERRNO_CONFLICTING_ATTRIBUTE_VALUES       = DRMAA_ERRNO_CONFLICTING_ATTRIBUTE_VALUES
ERRNO_TRY_LATER                          = DRMAA_ERRNO_TRY_LATER
ERRNO_DENIED_BY_DRM                      = DRMAA_ERRNO_DENIED_BY_DRM
ERRNO_INVALID_JOB                        = DRMAA_ERRNO_INVALID_JOB
ERRNO_RESUME_INCONSISTENT_STATE          = DRMAA_ERRNO_RESUME_INCONSISTENT_STATE
ERRNO_SUSPEND_INCONSISTENT_STATE         = DRMAA_ERRNO_SUSPEND_INCONSISTENT_STATE
ERRNO_HOLD_INCONSISTENT_STATE            = DRMAA_ERRNO_HOLD_INCONSISTENT_STATE
ERRNO_RELEASE_INCONSISTENT_STATE         = DRMAA_ERRNO_RELEASE_INCONSISTENT_STATE
ERRNO_EXIT_TIMEOUT                       = DRMAA_ERRNO_EXIT_TIMEOUT
ERRNO_NO_RUSAGE                          = DRMAA_ERRNO_NO_RUSAGE
ERRNO_NO_MORE_ELEMENTS                   = DRMAA_ERRNO_NO_MORE_ELEMENTS


cdef class drmaa_error( Exception ):
	""" Error returned by DRMAA. """
	property code:
		def __get__( self ):
			return self._code

	def __init__( self, int code, message = None ):
		if code == DRMAA_ERRNO_NO_MEMORY:
			raise MemoryError
		self._code = code
		if message is not None:
			self._message = str(message)
		else:
			self._message = drmaa_strerror( self._code )

	def __str__( self ):
		return "%s (code %d)" % (self._message, self._code)


class drmaa_assertion_failue( AssertionError ):
	def __init__( self, arg ):
		AssertionError.__init__( self, arg )


cdef class Session:
	"""
		DRMAA session.  Note that can be only one session object at once.
	"""

	cdef object __weakref__
	cdef int __opened
	cdef object __jobs
	cdef object __job_templates

	property opened:
		def __get__( self ):
			return bool( self.__opened )
	property jobs:
		def __get__( self ):
			return self.__jobs
	property job_templates:
		def __get__( self ):
			return self.__job_templates

	def __cinit__( self, contact=None, *args, **kw ):
		cdef char *c_contact
		if contact is None:
			c_contact = NULL
		else:
			contact = str(contact)
			c_contact = <char*>contact
		self.__opened = 0
		self.__jobs = []
		self.__job_templates = []
		_init_session( contact )
		self.__opened = 1

	def __dealloc__( self ):
		self.exit()

	def exit( self ):
		if self.__opened:
			_exit_session()
			self.__opened = 0

	def check_for_session( self ):
		if self.__opened == 0:
			raise drmaa_error( ERRNO_NO_ACTIVE_SESSION )

	def job_template( self, *argv, **attribs ):
		"""
			Constructs job template.
			Positional arguments (if any) are treaded as argv[] array
			while keywords arguments are used to initiate job attributes.
		"""
		self.check_for_session()
		return Job_template( self, *argv, **attribs )

	def get_attribute_names( self ):
		""" Returns list of scalar attribute names supported by DRMAA. """
		cdef char errmsg[ ERRLEN ]
		cdef drmaa_attr_names_t *list
		cdef int rc
		rc = drmaa_get_attribute_names( &list, errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )
		return convert_attr_names_list( list )

	def get_vector_attribute_names( self ):
		""" Returns list of vector attribute names supported by DRMAA. """
		cdef char errmsg[ ERRLEN ]
		cdef drmaa_attr_names_t *list
		cdef int rc
		rc = drmaa_get_vector_attribute_names( &list, errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )
		return convert_attr_names_list( list )

	def control( self, char *job_id, int action ):
		"""
			Control submitted job.
		"""
		cdef char errmsg[ ERRLEN ]
		cdef int rc
		self.check_for_session()
		rc = drmaa_control( job_id, action, errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )

	def job_ps( self, char *jobid ):
		""" Check status of submitted job. """
		cdef char errmsg[ ERRLEN ]
		cdef int remote_ps
		cdef int rc
		self.check_for_session()
		rc = drmaa_job_ps( jobid, &remote_ps, errmsg, ERRLEN )
		if rc:
			raise drmaa_error( rc, errmsg )
		else:
			return Job_state( remote_ps )

	def wait_for_any( self, timeout=None ):
		""" Wait for any job submitted within this session. """
		self.check_for_session()
		return self.wait( JOB_IDS_SESSION_ANY, timeout )

	def wait( self, char *jobid, timeout=None ):
		cdef char errmsg[ ERRLEN ]
		cdef size_t jobid_out_len
		cdef char jobid_out[ DRMAA_JOBNAME_BUFFER ]
		cdef int stat
		cdef drmaa_attr_values_t *rusage
		cdef int rc
		cdef int exited, aborted, signaled, coredumped, exitstatus
		cdef char signal[ DRMAA_SIGNAL_BUFFER ]

		self.check_for_session()
		if timeout is None:
			timeout = DRMAA_TIMEOUT_WAIT_FOREVER
		jobid_out_len = DRMAA_JOBNAME_BUFFER
		rc = drmaa_wait( jobid, jobid_out, jobid_out_len,
			&stat, timeout, &rusage, errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )
		resources = convert_attr_values_list(rusage)

		rc = drmaa_wifexited( &exited, stat, errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )
		rc = drmaa_wifaborted( &aborted, stat, errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )
		rc = drmaa_wifsignaled( &signaled, stat, errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )
		result = { "jobid": jobid_out, "rusage": resources }
		result[ "exited"   ] = exited
		result[ "aborted"  ] = aborted
		result[ "signaled" ] = signaled
		if exited:
			rc = drmaa_wexitstatus( &exitstatus, stat, errmsg, ERRLEN )
			if rc:  raise drmaa_error( rc, errmsg )
			result[ "exitstatus" ] = exitstatus
		if signaled:
			rc = drmaa_wtermsig( signal, DRMAA_SIGNAL_BUFFER, stat, errmsg, ERRLEN )
			if rc:  raise drmaa_error( rc, errmsg )
			result[ "signal"     ] = signal
			rc = drmaa_wcoredump( &coredumped, stat, errmsg, ERRLEN )
			if rc:  raise drmaa_error( rc, errmsg )
			result[ "coredumped" ] = coredumped
		return result

	def synchronize_all( self, timeout=None, dispose=True ):
		self.synchronize( JOB_IDS_SESSION_ALL, timeout, dispose )

	def synchronize( self, job_ids, timeout=None, dispose=True ):
		cdef char errmsg[ ERRLEN ]
		cdef int rc
		cdef char **list
		cdef size_t n_jobs, i
		cdef int timeout_arg

		self.check_for_session()
		n_jobs = len(job_ids)
		list = <char**>calloc( n_jobs+1, sizeof(char*) )
		j = iter(job_ids)
		for i from 0 <= i < n_jobs:
			tmp = j.next()
			list[i] = tmp
		list[n_jobs] = NULL
		if timeout is None:
			timeout_arg = DRMAA_TIMEOUT_WAIT_FOREVER
		else:
			timeout_arg = int(timeout)
		rc = drmaa_synchronize( list, timeout_arg, int(dispose),
				errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )

	def __str__( self ):
		return "%s: %s" % (self.DRM_system(), self.contact())

	def contact( self ):
		return _get_contact()

	def DRM_system( self ):
		return _get_DRM_system()


def version():
	cdef char errmsg[ ERRLEN ]
	cdef unsigned int major, minor
	cdef int rc
	rc = drmaa_version( &major, &minor, errmsg, ERRLEN )
	if rc:  raise drmaa_error( rc, errmsg )
	return (major, minor)

def contact():
	return _get_contact().split(",")

def DRM_system():
	return _get_DRM_system().split(",")



cdef class Job_template:
	"""
		Job template object.
	"""
	cdef object __weakref__
	cdef drmaa_job_template_t *jt
	cdef object session

	def __cinit__( self, Session session, *args, **kw ):
		cdef char errmsg[ ERRLEN ]
		cdef int rc
		if session is None:
			raise drmaa_error( DRMAA_ERRNO_NO_ACTIVE_SESSION )
		session.check_for_session()
		self.session = session
		self.session.job_templates.append( weakref.ref(self) )
		rc = drmaa_allocate_job_template( &self.jt, errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )

	def __init__( self, session, *argv, **attribs ):
		if len(argv):
			self.remote_command = argv[0]
			self.v_argv = argv[1:]
		for name, value in attribs.iteritems():
			setattr( self, name, value )

	def __dealloc__( self ):
		cdef char errmsg[ ERRLEN ]
		cdef int rc
		if self.jt != NULL and self.session.opened:
			rc = drmaa_delete_job_template( self.jt, errmsg, ERRLEN )
			if rc:  raise drmaa_error( rc, errmsg )
		self.session.job_templates.remove( weakref.ref(self) )
		self.session = None

	def __setattr__( self, char *name, value ):
		import types
		import __builtin__
		cdef char errmsg[ ERRLEN ]
		cdef char **list
		cdef int j,n,rc
		self.session.check_for_session()
		drmaa_name = "drmaa_" + name
		if type(value) in [ types.ListType, types.TupleType ]:
#print "vector", drmaa_name, repr(value)
			list_value = []
			for i in value:
				list_value.append( str(i) )
			n = len(list_value)
			list = <char**>calloc( n+1, sizeof(char*) )
			if list == NULL:
				raise MemoryError
			try:
				for i from 0 <= i < n:
					list[i] = list_value[i]
				rc = drmaa_set_vector_attribute( self.jt, drmaa_name, list,
						errmsg, ERRLEN )
				if rc == DRMAA_ERRNO_INVALID_ARGUMENT:
					raise AttributeError
				elif rc:
					raise drmaa_error( rc, errmsg )
			finally:
				free( list )
		else:
			str_value = str(value)
#print "scalar", drmaa_name, repr(value)
			rc = drmaa_set_attribute( self.jt, drmaa_name, str_value,
					errmsg, ERRLEN )
			if rc == DRMAA_ERRNO_INVALID_ARGUMENT:
				raise AttributeError
			elif rc:
				raise drmaa_error( rc, errmsg )

	def __getattr__( self, char *name ):
		cdef char errmsg[ ERRLEN ]
		cdef int rc
		cdef char value[ DRMAA_ATTR_BUFFER ]
		cdef drmaa_attr_values_t *values
		self.session.check_for_session()
		if name[:2] == "v_": # vector attribute
			rc = drmaa_get_vector_attribute( self.jt, name, &values, errmsg, ERRLEN )
			if rc == DRMAA_ERRNO_INVALID_ARGUMENT:
				raise AttributeError
			elif rc:
				raise drmaa_error( rc, errmsg )
			return convert_attr_values_list( values )
		else:
			rc = drmaa_get_attribute( self.jt, name, value, DRMAA_ATTR_BUFFER,
					errmsg, ERRLEN )
			if rc == DRMAA_ERRNO_INVALID_ARGUMENT:
				raise AttributeError
			elif rc:
				raise drmaa_error( rc, errmsg )
			return value

	def run( self ):
		"""
			Call drmaa_run_job for this template.
			Return job object.
		"""
		cdef char jobid[ DRMAA_JOBNAME_BUFFER ]
		cdef char errmsg[ ERRLEN ]
		cdef int rc
		self.session.check_for_session()
		rc = drmaa_run_job( jobid, DRMAA_JOBNAME_BUFFER, self.jt, errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )
		return Job( self.session, jobid )

	def run_bulk( self, int begin, int end, int incr=1 ):
		"""
			Call drmaa_run_bulk_jobs for this template.
			Return list of jobs.
		"""
		cdef drmaa_job_ids_t *jobids
		cdef char errmsg[ ERRLEN ]
		cdef int rc
		cdef size_t i,n
		cdef char jobid[ DRMAA_JOBNAME_BUFFER ]

		self.session.check_for_session()
		rc = drmaa_run_bulk_jobs( &jobids, self.jt, start, end, incr,
				errmsg, ERRLEN )
		if rc:  raise drmaa_error( rc, errmsg )
		try:
			rc = drmaa_get_num_job_ids( jobids, &n )
			if rc:  raise drmaa_error( rc )
			result = []
			for i from 0 <= i < n:
				rc = drmaa_get_next_job_id( jobids, jobid, DRMAA_JOBNAME_BUFFER )
				if rc:  raise drmaa_error( rc )
				result.append( Job( self.session, jobid ) )
		finally:
			drmaa_release_job_ids( jobids )
		return result



class Job( object ):
	"""
		Job submitted by DRMAA.
	"""
	def __init__( self, session, jobid ):
		session.check_for_session()
		self.session = session
		self.jobid   = jobid
		self.session.jobs.append( weakref.ref( self ) )

	def __dealloc__( self ):
		if self.session.opened:
			self.session.job_templates.remove( weakref.ref( self ) )
		self.session = None

	def wait( self, timeout=None ):
		"""
			Wait until job terminates.
		"""
		return self.session.wait( self.jobid, timeout )

	def control( self, action ):
		self.session.control( self.jobid, action )

	def hold( self ):
		"""
			Hold job in queue (will not run until released).
		"""
		self.control( DRMAA_CONTROL_HOLD )
	def release( self ):
		"""
			Release job.
		"""
		self.control( DRMAA_CONTROL_RELEASE )
	def suspend( self ):
		"""
			Suspend running job.
		"""
		self.control( DRMAA_CONTROL_SUSPEND )
	def resume( self ):
		"""
			Resume suspended job.
		"""
		self.control( DRMAA_CONTROL_RESUME )
	def terminate( self ):
		"""
			Terminate job.
		"""
		self.control( DRMAA_CONTROL_TERMINATE )

	def state( self ):
		"""
			Return state of job (drmaa_job_ps).
		"""
		return self.session.job_ps( self.jobid )

	def __str__( self ):
		return self.jobid
	def __cmp__( x, y ):
		return cmp( x.jobid, y.jobid )
	def __hash__( self ):
		return hash( self.jobid )

	def wait_for_state( self, end_state, prev_state=None, delay=1 ):
		import time
		assert prev_state != end_state
		end_states = to_job_state_list( end_state )
		prev_state = to_job_state( prev_state )
		min_order  = min( map( Job_state.order, end_states ) )
		while True:
			s = self.state()
			if s != prev_state and s.order() >= min_order:
				break
			#if s != prev_state and s.order() >= min_order:
			#	break
			time.sleep( delay )
		assert s in end_states, \
				"final state failure (is=%s, should be=%s, prev=%s)" \
				% (s, end_state, prev_state)
		return s



cdef class Job_state:
	"""
		Represents state of job.
	"""

	cdef int code

	def __cinit__( self, *args, **kw ):
		self.code = DRMAA_PS_UNDETERMINED

	def __init__( self, arg ):
		try:
			self.code = int(arg)
		except:
			self.code = {
				"undetermined":            DRMAA_PS_UNDETERMINED,
				"queued_active":           DRMAA_PS_QUEUED_ACTIVE,
				"system_on_hold":          DRMAA_PS_SYSTEM_ON_HOLD,
				"user_on_hold":            DRMAA_PS_USER_ON_HOLD,
				"user_system_on_hold":     DRMAA_PS_USER_SYSTEM_ON_HOLD,
				"running":                 DRMAA_PS_RUNNING,
				"system_suspended":        DRMAA_PS_SYSTEM_SUSPENDED,
				"user_suspended":          DRMAA_PS_USER_SUSPENDED,
				"user_system_suspended":   DRMAA_PS_USER_SYSTEM_SUSPENDED,
				"done":                    DRMAA_PS_DONE,
				"failed":                  DRMAA_PS_FAILED,
			} [ arg ]

	def __cmp__( Job_state self, Job_state rhs ):
		return int_cmp( self.code, rhs.code )

	def __hash__( self ):
		return self.code

	def order( self ):
		return {
			DRMAA_PS_UNDETERMINED:           0,
			DRMAA_PS_QUEUED_ACTIVE:          1,
			DRMAA_PS_SYSTEM_ON_HOLD:         1,
			DRMAA_PS_USER_ON_HOLD:           1,
			DRMAA_PS_USER_SYSTEM_ON_HOLD:    1,
			DRMAA_PS_RUNNING:                2,
			DRMAA_PS_SYSTEM_SUSPENDED:       2,
			DRMAA_PS_USER_SUSPENDED:         2,
			DRMAA_PS_USER_SYSTEM_SUSPENDED:  2,
			DRMAA_PS_DONE:                   3,
			DRMAA_PS_FAILED:                 3,
		} [ self.code ]

	def may_precede( self, next_state ):
		next_state = to_job_state( next_state )
		return self.order() <= next_state.order()

	def __str__( self ):
		return {
				DRMAA_PS_UNDETERMINED:           "undetermined",
				DRMAA_PS_QUEUED_ACTIVE:          "queued_active",
				DRMAA_PS_SYSTEM_ON_HOLD:         "system_on_hold",
				DRMAA_PS_USER_ON_HOLD:           "user_on_hold",
				DRMAA_PS_USER_SYSTEM_ON_HOLD:    "user_system_on_hold",
				DRMAA_PS_RUNNING:                "running",
				DRMAA_PS_SYSTEM_SUSPENDED:       "system_suspended",
				DRMAA_PS_USER_SUSPENDED:         "user_suspended",
				DRMAA_PS_USER_SYSTEM_SUSPENDED:  "user_system_suspended",
				DRMAA_PS_DONE:                   "done",
				DRMAA_PS_FAILED:                 "failed",
			} [ self.code ]

ps_undetermined          = Job_state( DRMAA_PS_UNDETERMINED )
ps_queued_active         = Job_state( DRMAA_PS_QUEUED_ACTIVE )
ps_system_on_hold        = Job_state( DRMAA_PS_SYSTEM_ON_HOLD )
ps_user_on_hold          = Job_state( DRMAA_PS_USER_ON_HOLD )
ps_user_system_on_hold   = Job_state( DRMAA_PS_USER_SYSTEM_ON_HOLD )
ps_running               = Job_state( DRMAA_PS_RUNNING )
ps_system_suspended      = Job_state( DRMAA_PS_SYSTEM_SUSPENDED )
ps_user_suspended        = Job_state( DRMAA_PS_USER_SUSPENDED )
ps_user_system_suspended = Job_state( DRMAA_PS_USER_SYSTEM_SUSPENDED )
ps_done                  = Job_state( DRMAA_PS_DONE )
ps_failed                = Job_state( DRMAA_PS_FAILED )


def to_job_state( arg ):
	if arg is None:
		return None
	elif isinstance( arg, Job_state ):
		return arg
	else:
		return Job_state( arg )


def to_job_state_list( arg ):
	if arg is None:
		return []
	elif isinstance( arg, Job_state ) \
				or type(arg) in ( type(str()), type(unicode()) ):
		return [ to_job_state(arg) ]
	else:
		return map( to_job_state, arg )


cdef int_cmp( int a, int b ):
	cdef int c
	c = a - b
	if c > 0:
		c = +1
	if c < 0:
		c = -1
	return c

cdef convert_attr_names_list( drmaa_attr_names_t *list ):
	cdef char errmsg[ ERRLEN ]
	cdef int rc
	cdef size_t i,n
	cdef char name[ DRMAA_ATTR_BUFFER ]
	try:
		rc = drmaa_get_num_attr_names( list, &n )
		if rc:  raise drmaa_error( rc )
		result = []
		for i from 0 <= i < n:
			rc = drmaa_get_next_attr_name( list, name, DRMAA_ATTR_BUFFER )
			if rc:  raise drmaa_error( rc )
			result.append( name )
	finally:
		drmaa_release_attr_names( list )
	return result


cdef convert_attr_values_list( drmaa_attr_values_t *list ):
	cdef char errmsg[ ERRLEN ]
	cdef int rc
	cdef size_t i,n
	cdef char value[ DRMAA_ATTR_BUFFER ]
	try:
		rc = drmaa_get_num_attr_values( list, &n )
		if rc:  raise drmaa_error( rc )
		result = []
		for i from 0 <= i < n:
			rc = drmaa_get_next_attr_value( list, value, DRMAA_ATTR_BUFFER )
			if rc:  raise drmaa_error( rc )
			result.append( value )
	finally:
		drmaa_release_attr_values( list )
	return result


def _init_session( contact=None ):
	cdef char errmsg[ ERRLEN ]
	cdef int rc
	cdef char *contact_str
	if contact is not None:
		contact_str = contact
	else:
		contact_str = NULL
	rc = drmaa_init( contact_str, errmsg, ERRLEN )
	if rc:  raise drmaa_error( rc, errmsg )


def _exit_session():
	cdef char errmsg[ ERRLEN ]
	cdef int rc
	rc = drmaa_exit( errmsg, ERRLEN )
	if rc:  raise drmaa_error( rc, errmsg )


def _get_contact():
	cdef int rc
	cdef char errmsg[ ERRLEN ]
	cdef char contact[ DRMAA_CONTACT_BUFFER ]
	rc = drmaa_get_contact( contact, DRMAA_CONTACT_BUFFER, errmsg, ERRLEN )
	if rc:  raise drmaa_error( rc, errmsg )
	return contact


def _get_DRM_system():
	cdef int rc
	cdef char errmsg[ ERRLEN ]
	cdef char drm_system[ DRMAA_DRM_SYSTEM_BUFFER ]
	rc = drmaa_get_DRM_system( drm_system, DRMAA_DRM_SYSTEM_BUFFER, errmsg, ERRLEN )
	if rc:  raise drmaa_error( rc, errmsg )
	return drm_system

